import unittest
from should_dsl import should
from fluidity import StateMachine, state, transition


class Door(StateMachine):
    state('closed')
    state('open')
    initial_state = 'closed'
    transition(from_='closed', event='open', to='open', action='open_action')
    transition(from_='open', event='close', to='closed', action='close_action')

    def open_action(self, when, where):
        self.when = when
        self.where = where

    def close_action(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class EventParameters(unittest.TestCase):

    def test_it_pass_parameters_received_by_event_to_action(self):
        door = Door()
        door.open('now!', 'there!')
        door |should| respond_to('when')
        door.when |should| equal_to('now!')
        door |should| respond_to('where')
        door.where |should| equal_to('there!')

    def test_it_pass_args_and_kwargs_to_action(self):
        door = Door()
        door.open('anytime', 'anywhere')
        door.close('1', 2, object, test=9, it=8, works=7)
        door |should| respond_to('args')
        door.args |should| equal_to(('1', 2, object))
        door |should| respond_to('kwargs')
        door.kwargs |should| equal_to({'test': 9, 'it': 8, 'works': 7})


if __name__ == '__main__':
    unittest.main()




from fluidity.backwardscompat import callable

# metaclass implementation idea from
# http://blog.ianbicking.org/more-on-python-metaprogramming-comment-14.html
_transition_gatherer = []

def transition(event, from_, to, action=None, guard=None):
    _transition_gatherer.append([event, from_, to, action, guard])

_state_gatherer = []

def state(name, enter=None, exit=None):
    _state_gatherer.append([name, enter, exit])


class MetaStateMachine(type):

    def __new__(cls, name, bases, dictionary):
        global _transition_gatherer, _state_gatherer
        Machine = super(MetaStateMachine, cls).__new__(cls, name, bases, dictionary)
        Machine._class_transitions = []
        Machine._class_states = {}
        for i in _transition_gatherer:
            Machine._add_class_transition(*i)
        for s in _state_gatherer:
            Machine._add_class_state(*s)
        _transition_gatherer = []
        _state_gatherer = []
        return Machine


StateMachineBase = MetaStateMachine('StateMachineBase', (object, ), {})


class StateMachine(StateMachineBase):

    def __init__(self):
        self.__class__._validate_machine_definitions()
        if callable(self.initial_state):
            self.initial_state = self.initial_state()
        self.current_state = self.initial_state
        self._handle_state_action(self.initial_state, 'enter')

    def __new__(cls, *args, **kwargs):
        obj = super(StateMachine, cls).__new__(cls)
        obj._states = {}
        obj._transitions = []
        return obj

    @classmethod
    def _validate_machine_definitions(cls):
        if not getattr(cls, '_class_states', None) or len(cls._class_states) < 2:
            raise InvalidConfiguration('There must be at least two states')
        if not getattr(cls, 'initial_state', None):
            raise InvalidConfiguration('There must exist an initial state')

    @classmethod
    def _add_class_state(cls, name, enter, exit):
        cls._class_states[name] = _State(name, enter, exit)

    def add_state(self, name, enter=None, exit=None):
        self._states[name] = _State(name, enter, exit)

    def states(self):
        return list(self.__class__._class_states.keys()) + list(self._states.keys())

    @classmethod
    def _add_class_transition(cls, event, from_, to, action, guard):
        cls._class_transitions.append(_Transition(event, from_, to, action, guard))
        this_event = cls._generate_event(event)
        setattr(cls, this_event.__name__, this_event)

    def add_transition(self, event, from_, to, action=None, guard=None):
        self._transitions.append(_Transition(event, from_, to, action, guard))
        this_event = self.__class__._generate_event(event)
        setattr(self, this_event.__name__,
            this_event.__get__(self, self.__class__))

    @classmethod
    def _generate_event(cls, name):
        def generated_event(self, *args, **kwargs):
            these_transitions = self._transitions_by_name(generated_event.__name__)
            these_transitions = self._ensure_from_validity(these_transitions)
            this_transition = self._check_guards(these_transitions)
            self._run_transition(this_transition, *args, **kwargs)
        generated_event.__doc__ = 'event %s' % name
        generated_event.__name__ = name
        return generated_event

    def _transitions_by_name(self, name):
        return filter(lambda transition: transition.event == name,
            self.__class__._class_transitions + self._transitions)

    def _ensure_from_validity(self, transitions):
        valid_transitions = []
        for transition in transitions:
            from_ = _listize(transition.from_)
            if self.current_state in from_:
                valid_transitions.append(transition)
        if len(valid_transitions) == 0:
            raise InvalidTransition("Cannot change from %s to %s" % (
                self.current_state, transition.to))
        return valid_transitions

    def _check_guards(self, transitions):
        allowed_transitions = []
        for transition in transitions:
            if self._check_guard(transition.guard):
                allowed_transitions.append(transition)
        if len(allowed_transitions) == 0:
            raise GuardNotSatisfied("Guard is not satisfied for this transition")
        elif len(allowed_transitions) > 1:
            raise ForkedTransition("More than one transition was allowed for this event")
        return allowed_transitions[0]

    def _run_transition(self, transition, *args, **kwargs):
        self._handle_state_action(self.current_state, 'exit')
        self.current_state = transition.to
        self._handle_state_action(transition.to, 'enter')
        self._handle_action(transition.action, *args, **kwargs)

    def _handle_state_action(self, state, kind):
        try:
            action = getattr(self._class_states[state], kind)
        except KeyError:
            action = getattr(self._states[state], kind)
        self._run_action_or_list(action)

    def _handle_action(self, action, *args, **kwargs):
        self._run_action_or_list(action, *args, **kwargs)

    def _run_action_or_list(self, action_param, *args, **kwargs):
        if not action_param:
            return
        action_items = _listize(action_param)
        for action_item in action_items:
            self._run_action(action_item, *args, **kwargs)

    def _run_action(self, action, *args, **kwargs):
        if callable(action):
            self._try_to_run_with_args(action, self, *args, **kwargs)
        else:
            self._try_to_run_with_args(getattr(self, action), *args, **kwargs)

    def _try_to_run_with_args(self, action, *args, **kwargs):
        try:
            action(*args, **kwargs)
        except TypeError:
            if len(args) > 0 and args[0] == self:
                action(self)
            else:
                action()

    def _check_guard(self, guard_param):
        if guard_param is None:
            return True
        guard_items = _listize(guard_param)
        result = True
        for guard_item in guard_items:
            result = result and self._evaluate_guard(guard_item)
        return result

    def _evaluate_guard(self, guard):
        if callable(guard):
            return guard(self)
        else:
            guard = getattr(self, guard)
            if callable(guard):
                guard = guard()
            return guard


class _Transition(object):

    def __init__(self, event, from_, to, action, guard):
        self.event = event
        self.from_ = from_
        self.to = to
        self.action = action
        self.guard = guard


class _State(object):

    def __init__(self, name, enter, exit):
        self.name = name
        self.enter = enter
        self.exit = exit


class InvalidConfiguration(Exception):
    pass


class InvalidTransition(Exception):
    pass


class GuardNotSatisfied(Exception):
    pass


class ForkedTransition(Exception):
    pass


def _listize(value):
    return type(value) in [list, tuple] and value or [value]




import unittest
from should_dsl import should
from fluidity import StateMachine, state, transition


class CrazyGuy(StateMachine):
    state('looking', exit='no_lookin_anymore')
    state('falling', enter='will_fall_right_now')
    initial_state = 'looking'
    transition(from_='looking', event='jump', to='falling',
               action='become_at_risk', guard='always_can_jump')

    def __init__(self):
        StateMachine.__init__(self)
        self.at_risk = False
        self.callbacks = []

    def become_at_risk(self):
        self.at_risk = True
        self.callbacks.append('action')

    def no_lookin_anymore(self):
        self.callbacks.append('old exit')

    def will_fall_right_now(self):
        self.callbacks.append('new enter')

    def always_can_jump(self):
        self.callbacks.append('guard')
        return True


class CallbackOrder(unittest.TestCase):

    def setUp(self):
        guy = CrazyGuy()
        guy.jump()
        self.callbacks = guy.callbacks

    def test_it_runs_guard_first(self):
        '''(1) guard'''
        self.callbacks[0] |should| equal_to('guard')

    def test_it_and_then_old_state_exit(self):
        '''(2) old state exit action'''
        self.callbacks[1] |should| equal_to('old exit')

    def test_it_and_then_new_state_exit(self):
        '''(3) new state enter action'''
        self.callbacks[2] |should| equal_to('new enter')

    def test_it_and_then_transaction_action(self):
        '''(4) transaction action'''
        self.callbacks[3] |should| equal_to('action')


if __name__ == '__main__':
    unittest.main()




import unittest
from should_dsl import should, should_not
from fluidity import StateMachine, state, transition, ForkedTransition


class LoanRequest(StateMachine):
    state('pending')
    state('analyzing')
    state('refused')
    state('accepted')
    initial_state = 'pending'
    transition(from_='pending', event='analyze', to='analyzing', action='input_data')
    transition(from_='analyzing', event='forward_analysis_result',
               guard='was_loan_accepted', to='accepted')
    transition(from_='analyzing', event='forward_analysis_result',
               guard='was_loan_refused', to='refused')

    def input_data(self, accepted=True):
        self.accepted = accepted

    def was_loan_accepted(self):
        return self.accepted or getattr(self, 'truify', False)

    def was_loan_refused(self):
        return not self.accepted or getattr(self, 'truify', False)


class FluidityEventSupportsMultipleTransitions(unittest.TestCase):
    '''Event chooses one of multiple transitions, based in their guards'''

    def test_it_selects_the_transition_having_a_passing_guard(self):
        request = LoanRequest()
        request.analyze()
        request.forward_analysis_result()
        request.current_state |should| equal_to('accepted')

        request = LoanRequest()
        request.analyze(accepted=False)
        request.forward_analysis_result()
        request.current_state |should| equal_to('refused')

    def test_it_raises_error_if_more_than_one_guard_passes(self):
        request = LoanRequest()
        request.analyze()
        request.truify = True
        request.forward_analysis_result |should| throw(
            ForkedTransition, message="More than one transition was allowed for this event")



if __name__ == '__main__':
    unittest.main()




import unittest
from should_dsl import should, should_not
from fluidity import StateMachine, state, transition


class Door(StateMachine):
    state('closed')
    state('open')
    initial_state = 'closed'
    transition(from_='closed', event='open', to='open')


class IndividuationSpec(unittest.TestCase):
    '''Fluidity object (individuation)'''

    def setUp(self):
        self.door = Door()
        self.door.add_state('broken')
        self.door.add_transition(from_='closed', event='crack', to='broken')

    def test_it_responds_to_an_event(self):
        self.door |should| respond_to('crack')

    def test_its_event_changes_its_state_when_called(self):
        self.door.crack()
        self.door.current_state |should| equal_to('broken')

    def test_it_informs_all_its_states(self):
        self.door |should| have(3).states
        self.door.states() |should| include_all_of(['open', 'closed', 'broken'])

    def test_its_individuation_does_not_affect_other_objects_from_the_same_class(self):
        another_door = Door()
        another_door |should_not| respond_to('crack')


if __name__ == '__main__':
    unittest.main()




import unittest
from should_dsl import should, should_not
from fluidity import StateMachine, state, transition
from fluidity import InvalidTransition


class MyMachine(StateMachine):

    initial_state = 'created'

    state('created')
    state('waiting')
    state('processed')
    state('canceled')

    transition(from_='created', event='queue', to='waiting')
    transition(from_='waiting', event='process', to='processed')
    transition(from_=['waiting', 'created'], event='cancel', to='canceled')


class FluidityEvent(unittest.TestCase):

    def test_its_declaration_creates_a_method_with_its_name(self):
        machine = MyMachine()
        machine |should| respond_to('queue')
        machine |should| respond_to('process')

    def test_it_changes_machine_state(self):
        machine = MyMachine()
        machine.current_state |should| equal_to('created')
        machine.queue()
        machine.current_state |should| equal_to('waiting')
        machine.process()
        machine.current_state |should| equal_to('processed')

    def test_it_ensures_event_order(self):
        machine = MyMachine()
        machine.process |should| throw(InvalidTransition)
        machine.queue()
        machine.queue |should| throw(InvalidTransition)
        machine.process |should_not| throw(Exception)

    def test_it_accepts_multiple_origin_states(self):
        machine = MyMachine()
        machine.cancel |should_not| throw(Exception)

        machine = MyMachine()
        machine.queue()
        machine.cancel |should_not| throw(Exception)

        machine = MyMachine()
        machine.queue()
        machine.process()
        machine.cancel |should| throw(Exception)


if __name__ == '__main__':
    unittest.main()




import unittest
import time
from should_dsl import should
from fluidity import StateMachine, transition, state


class FluidityState(unittest.TestCase):

    def test_it_defines_states(self):
        class MyMachine(StateMachine):
            state('unread')
            state('read')
            state('closed')
            initial_state = 'read'
        machine = MyMachine()
        machine |should| have(3).states
        machine.states() |should| include_all_of(['unread', 'read', 'closed'])

    def test_it_has_an_initial_state(self):
        class MyMachine(StateMachine):
            initial_state = 'closed'
            state('open')
            state('closed')
        machine = MyMachine()
        machine.initial_state |should| equal_to('closed')
        machine.current_state |should| equal_to('closed')

    def test_it_defines_states_using_method_calls(self):
        class MyMachine(StateMachine):
            state('unread')
            state('read')
            state('closed')
            initial_state = 'unread'
            transition(from_='unread', event='read', to='read')
            transition(from_='read', event='close', to='closed')
        machine = MyMachine()
        machine |should| have(3).states
        machine.states() |should| include_all_of(['unread', 'read', 'closed'])

        class OtherMachine(StateMachine):
            state('idle')
            state('working')
            initial_state = 'idle'
            transition(from_='idle', event='work', to='working')
        machine = OtherMachine()
        machine |should| have(2).states
        machine.states() |should| include_all_of(['idle', 'working'])

    def test_its_initial_state_may_be_a_callable(self):
        def is_business_hours():
            return True
        class Person(StateMachine):
            initial_state = lambda person: (person.worker and is_business_hours()) and 'awake' or 'sleeping'
            state('awake')
            state('sleeping')
            def __init__(self, worker):
                self.worker = worker
                StateMachine.__init__(self)

        person = Person(worker=True)
        person.current_state |should| equal_to('awake')

        person = Person(worker=False)
        person.current_state |should| equal_to('sleeping')


if __name__ == '__main__':
    unittest.main()




import unittest
from should_dsl import should, should_not
from fluidity import StateMachine, state, transition


class CrazyGuy(StateMachine):
    state('looking')
    state('falling')
    initial_state = 'looking'
    transition(from_='looking', event='jump', to='falling', action=['become_at_risk', 'accelerate'])

    def __init__(self):
        StateMachine.__init__(self)
        self.at_risk = False
        self.accelerating = False

    def become_at_risk(self):
        self.at_risk = True

    def accelerate(self):
        self.accelerating = True


class FluidityTransitionAction(unittest.TestCase):

    def test_it_runs_when_transition_occurs(self):
        guy = CrazyGuy()
        guy |should_not| be_at_risk
        guy.jump()
        guy |should| be_at_risk

    def test_it_supports_multiple_transition_actions(self):
        guy = CrazyGuy()
        guy |should_not| be_at_risk
        guy |should_not| be_accelerating
        guy.jump()
        guy |should| be_at_risk
        guy |should| be_accelerating


if __name__ == '__main__':
    unittest.main()




import unittest
from should_dsl import should
from fluidity import StateMachine, state, transition
from fluidity import GuardNotSatisfied


class JumperGuy(StateMachine):
    state('looking', enter=lambda jumper: jumper.footsteps.append('enter looking'),
                     exit=lambda jumper: jumper.footsteps.append('exit looking'))
    state('falling', enter=lambda jumper: jumper.footsteps.append('enter falling'))
    initial_state = 'looking'

    transition(from_='looking', event='jump', to='falling',
               action=lambda jumper: jumper.footsteps.append('action jump'),
               guard=lambda jumper: jumper.footsteps.append('guard jump') is None)

    def __init__(self):
        self.footsteps = []
        StateMachine.__init__(self)


class CallableSupport(unittest.TestCase):

    def test_every_callback_can_be_a_callable(self):
        '''every callback can be a callable'''
        guy = JumperGuy()
        guy.jump()
        guy |should| have(5).footsteps
        guy.footsteps |should| include_all_of([
            'enter looking', 'exit looking', 'enter falling',
            'action jump', 'guard jump'])

    def test_it_should_deny_state_change_if_guard_callable_returns_false(self):
        class Door(StateMachine):
            state('open')
            state('closed')
            initial_state = 'closed'
            transition(from_='closed', event='open', to='open',
                       guard=lambda d: not door.locked)
            def locked(self):
                return self.locked

        door = Door()
        door.locked = True
        door.open |should| throw(GuardNotSatisfied)


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

version = '0.2.0'
readme = open('README.rst').read()

setup(name='fluidity-sm',
      version=version,
      description='Fluidity: state machine implementation for Python objects',
      long_description=readme,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.4',
          'Programming Language :: Python :: 2.5',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.1',
          'Programming Language :: Python :: 3.2',
          'Topic :: Software Development :: Libraries',
      ],
      keywords='state machine python dsl',
      author='Rodrigo Manhães',
      author_email='rmanhaes@gmail.com',
      url='https://github.com/nsi-iff/fluidity',
      license='MIT License',
      packages=find_packages()
      )




import unittest
from should_dsl import should, should_not
from fluidity import StateMachine, state, transition
from fluidity import GuardNotSatisfied


class FallingMachine(StateMachine):
    state('looking')
    state('falling')
    initial_state = 'looking'
    transition(from_='looking', event='jump', to='falling',
               guard=['ready_to_fly', 'high_enough'])

    def __init__(self, ready=True):
        StateMachine.__init__(self)
        self.ready = ready
        self.high_enough_flag = True

    def ready_to_fly(self):
        return self.ready

    def high_enough(self):
        return self.high_enough_flag


class FluidityGuard(unittest.TestCase):

    def test_it_allows_transition_if_satisfied(self):
        machine = FallingMachine()
        machine.jump |should_not| throw(Exception)
        machine.current_state |should| equal_to('falling')

    def test_it_forbids_transition_if_not_satisfied(self):
        machine = FallingMachine(ready=False)
        machine.jump |should| throw(GuardNotSatisfied)

    def test_it_may_be_an_attribute(self):
        '''it may be an attribute, not only a method'''
        machine = FallingMachine()
        machine.ready_to_fly = False
        machine.jump |should| throw(GuardNotSatisfied)

        machine.ready_to_fly = True
        machine.jump |should_not| throw(Exception)
        machine.current_state |should| equal_to('falling')

    def test_it_allows_transition_only_if_all_are_satisfied(self):
        machine = FallingMachine()
        machine.ready_to_fly = True
        machine.high_enough_flag = True
        machine.jump |should_not| throw(Exception)

        machine = FallingMachine()
        machine.ready_to_fly = False
        machine.high_enough_flag = True
        machine.jump |should| throw(GuardNotSatisfied)

        machine = FallingMachine()
        machine.ready_to_fly = True
        machine.high_enough_flag = False
        machine.jump |should| throw(GuardNotSatisfied)


if __name__ == '__main__':
    unittest.main()




import unittest
from should_dsl import should, should_not
from fluidity import StateMachine, state, transition


class ActionMachine(StateMachine):

    state('created', enter='about_to_create',
                     exit=['other_post_create', 'post_create'])
    state('waiting', enter=['pre_wait', 'other_pre_wait'])
    initial_state = 'created'
    transition(from_='created', event='queue', to='waiting')

    def __init__(self):
        self.enter_create = False
        super(ActionMachine, self).__init__()
        self.is_enter_aware = False
        self.is_exit_aware = False
        self.pre_wait_aware = False
        self.other_pre_wait_aware = False
        self.post_create_aware = False
        self.other_post_create_aware = False
        self.count = 0

    def pre_wait(self):
        self.is_enter_aware = True
        self.pre_wait_aware = True
        if getattr(self, 'pre_wait_expectation', None):
            self.pre_wait_expectation()

    def post_create(self):
        self.is_exit_aware = True
        self.post_create_aware = True
        if getattr(self, 'post_create_expectation', None):
            self.post_create_expectation()

    def about_to_create(self):
        self.enter_create = True

    def other_pre_wait(self):
        self.other_pre_wait_aware = True

    def other_post_create(self):
        self.other_post_create_aware = True



class FluidityAction(unittest.TestCase):

    def test_it_runs_enter_action_before_machine_enters_a_given_state(self):
        machine = ActionMachine()
        machine |should_not| be_enter_aware
        machine.queue()
        machine |should| be_enter_aware

    def test_it_runs_exit_action_after_machine_exits_a_given_state(self):
        machine = ActionMachine()
        machine |should_not| be_exit_aware
        machine.queue()
        machine |should| be_enter_aware

    def test_it_runs_exit_action_before_enter_action(self):
        '''it runs old state's exit action before new state's enter action'''
        machine = ActionMachine()
        def post_create_expectation(_self):
            _self.count +=1
            _self.count |should| be(1)
        def pre_wait_expectation(_self):
            _self.count += 1
            _self.count |should| be(2)
        machine.post_create_expectation = \
            post_create_expectation.__get__(machine, ActionMachine)
        machine.pre_wait_expectation = \
              pre_wait_expectation.__get__(machine, ActionMachine)
        machine.queue()

    def test_it_runs_enter_action_for_initial_state_at_creation(self):
        ActionMachine().enter_create |should| be(True)

    def test_it_accepts_many_enter_actions(self):
        machine = ActionMachine()
        machine |should_not| be_pre_wait_aware
        machine |should_not| be_other_pre_wait_aware
        machine.queue()
        machine |should| be_pre_wait_aware
        machine |should| be_other_pre_wait_aware

    def test_it_accepts_exit_actions(self):
        machine = ActionMachine()
        machine |should_not| be_post_create_aware
        machine |should_not| be_other_post_create_aware
        machine.queue()
        machine |should| be_post_create_aware
        machine |should| be_other_post_create_aware


if __name__ == '__main__':
    unittest.main()




import unittest
from should_dsl import should
from fluidity import StateMachine, state, InvalidConfiguration


class FluidityConfigurationValidation(unittest.TestCase):

    def test_it_requires_at_least_two_states(self):
        class MyMachine(StateMachine):
            pass
        MyMachine |should| throw(InvalidConfiguration,
            message="There must be at least two states")
        class OtherMachine(StateMachine):
            state('open')
        OtherMachine |should| throw(InvalidConfiguration,
            message="There must be at least two states")

    def test_it_requires_an_initial_state(self):
        class MyMachine(StateMachine):
            state('open')
            state('closed')
        MyMachine |should| throw(InvalidConfiguration,
            message="There must exist an initial state")
        class AnotherMachine(StateMachine):
            state('open')
            state('closed')
            initial_state = None
        AnotherMachine |should| throw(InvalidConfiguration,
            message="There must exist an initial state")


if __name__ == '__main__':
    unittest.main()

